@@ -38,17 +38,12 @@ static void pyiClass(sipSpec *pt, moduleDef *mod, classDef *cd,
         ifaceFileList **defined, int indent, FILE *fp);
 static void pyiMappedType(sipSpec *pt, moduleDef *mod, mappedTypeDef *mtd,
         ifaceFileList **defined, int indent, FILE *fp);
-static void pyiCtor(sipSpec *pt, moduleDef *mod, classDef *cd, ctorDef *ct,
-        int overloaded, ifaceFileList *defined, int indent, FILE *fp);
 static void pyiCallable(sipSpec *pt, moduleDef *mod, memberDef *md,
         overDef *overloads, int is_method, ifaceFileList *defined, int indent,
         FILE *fp);
 static void pyiProperty(sipSpec *pt, moduleDef *mod, propertyDef *pd,
         int is_setter, memberDef *md, overDef *overloads,
         ifaceFileList *defined, int indent, FILE *fp);
-static void pyiOverload(sipSpec *pt, moduleDef *mod, overDef *od,
-        int overloaded, int is_method, ifaceFileList *defined, int indent,
-        int pep484, FILE *fp);
 static void pyiPythonSignature(sipSpec *pt, moduleDef *mod, signatureDef *sd,
         int need_self, ifaceFileList *defined, KwArgs kwargs, int pep484,
         FILE *fp);
@@ -86,18 +81,19 @@ static typeHintNodeDef *flatten_unions(typeHintNodeDef *nodes);
 static typeHintNodeDef *copyTypeHintNode(sipSpec *pt, typeHintDef *thd,
         int out);
 static int isPyKeyword(const char *word);
+static void appendToIfaceFileList(ifaceFileList **ifflp, ifaceFileDef *iff);
 
 
 /*
  * Generate the .pyi file.
  */
-void generateTypeHints(sipSpec *pt, moduleDef *mod, const char *pyiFile)
+int generateTypeHints(sipSpec *pt, moduleDef *mod, const char *pyiFile)
 {
     FILE *fp;
 
     /* Generate the file. */
     if ((fp = fopen(pyiFile, "w")) == NULL)
-        fatal("Unable to create file \"%s\"\n", pyiFile);
+        return error("Unable to create file \"%s\"\n", pyiFile);
 
     /* Write the header. */
     fprintf(fp,
@@ -123,6 +119,8 @@ void generateTypeHints(sipSpec *pt, moduleDef *mod, const char *pyiFile)
         pyiModule(pt, mod, fp);
 
     fclose(fp);
+
+    return 0;
 }
 
 
@@ -309,7 +307,14 @@ static void pyiClass(sipSpec *pt, moduleDef *mod, classDef *cd,
         }
         else if (cd->supertype != NULL)
         {
-            fprintf(fp, "%s", cd->supertype->text);
+            /*
+             * This is a hack to correct the .pyi file when the supertype
+             * hasn't been given a fully qualified name (ie. with ABI v12).
+             */
+            if (sipName != NULL && strncmp(cd->supertype->text, "sip.", 4) == 0)
+                fprintf(fp, "%s.%s", sipName, &cd->supertype->text[4]);
+            else
+                fprintf(fp, "%s", cd->supertype->text);
         }
         else
         {
@@ -507,7 +512,7 @@ static void pyiMappedType(sipSpec *pt, moduleDef *mod, mappedTypeDef *mtd,
     {
         separate(TRUE, indent, fp);
         prIndent(indent, fp);
-        fprintf(fp, "class %s(sip.wrapper):\n", mtd->pyname->text);
+        fprintf(fp, "class %s(%s.wrapper):\n", mtd->pyname->text, (sipName != NULL ? sipName : "sip"));
 
         ++indent;
 
@@ -531,19 +536,10 @@ static void pyiMappedType(sipSpec *pt, moduleDef *mod, mappedTypeDef *mtd,
 }
 
 
-/*
- * Generate a ctor docstring.
- */
-void dsCtor(sipSpec *pt, classDef *cd, ctorDef *ct, FILE *fp)
-{
-    pyiCtor(pt, pt->module, cd, ct, FALSE, NULL, 0, fp);
-}
-
-
 /*
  * Generate an ctor type hint.
  */
-static void pyiCtor(sipSpec *pt, moduleDef *mod, classDef *cd, ctorDef *ct,
+void pyiCtor(sipSpec *pt, moduleDef *mod, classDef *cd, ctorDef *ct,
         int overloaded, ifaceFileList *defined, int indent, FILE *fp)
 {
     int a, need_comma;
@@ -773,24 +769,13 @@ static void pyiProperty(sipSpec *pt, moduleDef *mod, propertyDef *pd,
 }
 
 
-/*
- * Generate the docstring for a single API overload.
- */
-void dsOverload(sipSpec *pt, overDef *od, int is_method, FILE *fp)
-{
-    pyiOverload(pt, pt->module, od, FALSE, is_method, NULL, 0, FALSE, fp);
-}
-
-
 /*
  * Generate the type hints for a single API overload.
  */
-static void pyiOverload(sipSpec *pt, moduleDef *mod, overDef *od,
-        int overloaded, int is_method, ifaceFileList *defined,
-        int indent, int pep484, FILE *fp)
+void pyiOverload(sipSpec *pt, moduleDef *mod, overDef *od, int overloaded,
+        int is_method, ifaceFileList *defined, int indent, int pep484,
+        FILE *fp)
 {
-    int need_self;
-
     if (overloaded)
     {
         prIndent(indent, fp);
@@ -806,10 +791,18 @@ static void pyiOverload(sipSpec *pt, moduleDef *mod, overDef *od,
     prIndent(indent, fp);
     fprintf(fp, "%s%s", (pep484 ? "def " : ""), od->common->pyname->text);
 
-    need_self = (is_method && !isStatic(od));
+    if (pep484 && (od->common->slot == eq_slot || od->common->slot == ne_slot))
+    {
+        /* mypy recommends using 'object' as the argument type. */
+        fprintf(fp, "(self, other: object)");
+    }
+    else
+    {
+        int need_self = (is_method && !isStatic(od));
 
-    pyiPythonSignature(pt, mod, &od->pysig, need_self, defined, od->kwargs,
-            pep484, fp);
+        pyiPythonSignature(pt, mod, &od->pysig, need_self, defined, od->kwargs,
+                pep484, fp);
+    }
 
     if (pep484)
         fprintf(fp, ": ...\n");
@@ -861,6 +854,9 @@ static int pyiArgument(sipSpec *pt, moduleDef *mod, argDef *ad, int arg_nr,
         }
     }
 
+    if (isArray(ad))
+        fprintf(fp, "%s.array[", (sipName != NULL) ? sipName : "sip");
+
     pyiType(pt, mod, ad, out, defined, pep484, fp);
 
     if (names && ad->atype == ellipsis_type)
@@ -872,6 +868,9 @@ static int pyiArgument(sipSpec *pt, moduleDef *mod, argDef *ad, int arg_nr,
             fprintf(fp, "a%d", arg_nr);
     }
 
+    if (isArray(ad))
+        fprintf(fp, "]");
+
     if (optional)
     {
         if (use_optional)
@@ -1323,20 +1322,6 @@ static int inIfaceFileList(ifaceFileDef *iff, ifaceFileList *defined)
 }
 
 
-/*
- * Create a new type hint for a raw string.
- */
-typeHintDef *newTypeHint(char *raw_hint)
-{
-    typeHintDef *thd = sipMalloc(sizeof (typeHintDef));
-
-    thd->status = needs_parsing;
-    thd->raw_hint = raw_hint;
-
-    return thd;
-}
-
-
 /*
  * Generate a type hint from a /TypeHint/ annotation.
  */
@@ -1905,3 +1890,54 @@ static int isPyKeyword(const char *word)
 
     return FALSE;
 }
+
+
+/*
+ * Add an interface file to an interface file list if it isn't already there.
+ */
+static void appendToIfaceFileList(ifaceFileList **ifflp, ifaceFileDef *iff)
+{
+    /* Make sure we don't try to add an interface file to its own list. */
+    if (&iff->used != ifflp)
+    {
+        ifaceFileList *iffl;
+
+        while ((iffl = *ifflp) != NULL)
+        {
+            /* Don't bother if it is already there. */
+            if (iffl->iff == iff)
+                return;
+
+            ifflp = &iffl -> next;
+        }
+
+        iffl = sipMalloc(sizeof (ifaceFileList));
+
+        iffl->iff = iff;
+        iffl->next = NULL;
+
+        *ifflp = iffl;
+    }
+}
+
+
+/*
+ * Generate a fully qualified class name as a reST reference.
+ */
+void restPyClass(classDef *cd, FILE *fp)
+{
+    fprintf(fp, ":sip:ref:`~%s.", cd->iff->module->fullname->text);
+    prScopedPythonName(fp, cd->ecd, cd->pyname->text);
+    fprintf(fp, "`");
+}
+
+
+/*
+ * Generate a fully qualified enum name as a reST reference.
+ */
+void restPyEnum(enumDef *ed, FILE *fp)
+{
+    fprintf(fp, ":sip:ref:`~%s.", ed->module->fullname->text);
+    prScopedPythonName(fp, ed->ecd, ed->pyname->text);
+    fprintf(fp, "`");
+}
